@@ -2,33 +2,38 @@
 
 """Field definitions for the flexible_forms module."""
 
+import json
 import logging
+import urllib.parse as urlparse
 from typing import (
+    Set,
     TYPE_CHECKING,
     Any,
     Dict,
     List,
-    Mapping,
     Optional,
     Sequence,
     Tuple,
     Type,
+    TypedDict,
     cast,
 )
 
-import jmespath
 import requests
 import simpleeval
+from django.apps import apps
 from django.core.paginator import Paginator
 from django.db.models import FileField, ImageField
-from django.db.models import fields as model_fields
+from django.db.models import Model as DjangoModel, QuerySet, fields as model_fields
 from django.forms import fields as form_fields
 from django.forms import widgets as form_widgets
-from django.http.request import HttpRequest
+from django.forms import model_to_dict
+from django.http import HttpRequest, QueryDict
 from django.shortcuts import get_object_or_404
 from django.template import Context, Template
 from django.template.base import VariableNode
-from django.urls import reverse
+from django.urls import Resolver404, resolve, reverse
+from django.utils.functional import cached_property
 
 from flexible_forms.widgets import (
     AutocompleteResult,
@@ -43,7 +48,12 @@ except ImportError:  # pragma: no cover
     from django.contrib.postgres.fields import JSONField
     from django.contrib.postgres.forms import JSONField as JSONFormField
 
-from flexible_forms.utils import evaluate_expression, stable_json
+from flexible_forms.utils import (
+    evaluate_expression,
+    get_expression_fields,
+    jp,
+    stable_json,
+)
 
 logger = logging.getLogger(__name__)
 
@@ -200,10 +210,10 @@ class FieldType(metaclass=FieldTypeMetaclass):
         cls,
         *,
         field: "BaseField",
+        field_values: Dict[str, Optional[Any]],
         record: Optional["BaseRecord"] = None,
         modifiers: Sequence[Tuple[str, str]] = (),
-        field_values: Optional[Mapping[str, Optional[Any]]] = None,
-        **kwargs: Any,
+        **form_field_options: Any,
     ) -> form_fields.Field:
         """Return an instance of the field for use in a Django form.
 
@@ -215,13 +225,13 @@ class FieldType(metaclass=FieldTypeMetaclass):
                 available.
             modifiers: A sequence of modifiers to be applied to the field.
             field_values: A mapping of the current form values.
-            kwargs: Passed to the form field constructor.
+            form_field_options: Passed to the form field constructor.
 
         Returns:
             form_fields.Field: An instance of the form field.
         """
         # If no widget was given explicitly, build a default one.
-        kwargs["widget"] = kwargs.get(
+        form_field_options["widget"] = form_field_options.get(
             "widget", cls.as_form_widget(field=field, record=record)
         )
 
@@ -229,13 +239,15 @@ class FieldType(metaclass=FieldTypeMetaclass):
         form_field = cls.form_field_class(
             **{
                 **cls.form_field_options,
-                **kwargs,
+                **form_field_options,
             },
         )
 
         # Apply any modifiers to the field.
         form_field = cls.apply_modifiers(
-            form_field=form_field,
+            form_field,
+            field=field,
+            record=record,
             modifiers=modifiers,
             field_values=field_values,
         )
@@ -244,7 +256,11 @@ class FieldType(metaclass=FieldTypeMetaclass):
 
     @classmethod
     def as_form_widget(
-        cls, field: "BaseField", record: Optional["BaseRecord"] = None, **kwargs: Any
+        cls,
+        *,
+        field: "BaseField",
+        record: Optional["BaseRecord"] = None,
+        **form_widget_options: Any,
     ) -> form_widgets.Widget:
         """Return an instance of the form widget for rendering.
 
@@ -254,7 +270,7 @@ class FieldType(metaclass=FieldTypeMetaclass):
             field: The Field model instance.
             record: The record instance to which the form is bound, if
                 available.
-            kwargs: Passed through to the form widget constructor.
+            form_widget_options: Passed through to the form widget constructor.
 
         Returns:
             form_widgets.Widget: The configured form widget for the field.
@@ -264,19 +280,19 @@ class FieldType(metaclass=FieldTypeMetaclass):
         return widget_cls(
             **{
                 **cls.form_widget_options,
-                **kwargs,
+                **form_widget_options,
             }
         )
 
     @classmethod
-    def as_model_field(cls, **kwargs: Any) -> model_fields.Field:
+    def as_model_field(cls, **model_field_options: Any) -> model_fields.Field:
         """Return an instance of the field for use in a Django model.
 
         Receives a dict of kwargs to pass through to the model field constructor.
 
         Args:
-            kwargs: A dict of kwargs to be passed to the constructor of the
-                `model_field_class`.
+            model_field_options: A dict of kwargs to be passed to the
+                constructor of the `model_field_class`.
 
         Returns:
             model_fields.Field: An instance of the model field.
@@ -284,7 +300,7 @@ class FieldType(metaclass=FieldTypeMetaclass):
         return cls.model_field_class(
             **{
                 **cls.model_field_options,
-                **kwargs,
+                **model_field_options,
             }
         )
 
@@ -292,27 +308,41 @@ class FieldType(metaclass=FieldTypeMetaclass):
     def apply_modifiers(
         cls,
         form_field: form_fields.Field,
+        *,
+        field: "BaseField",
+        field_values: Dict[str, Any],
+        record: Optional["BaseRecord"] = None,
         modifiers: Sequence[Tuple[str, str]] = (),
-        field_values: Optional[Mapping[str, Any]] = None,
     ) -> form_fields.Field:
         """Apply the given modifiers to the given Django form field.
 
         Args:
             form_field: The form field to be modified.
+            field: The BaseField instance.
+            field_values: The current values of all fields on the form.
+            record: The BaseRecord instance, if available.
             modifiers: A sequence of tuples
                 in the form of (attribute, expression) tuples to apply to the
                 field.
-            field_values: The current values of
-                all fields on the form.
 
         Returns:
             form_fields.Field: The given form field, modified using the modifiers.
         """
         for attribute, expression in modifiers:
+            expression_context = field_values
+
+            if record:
+                record_variable = (
+                    (record._meta.verbose_name or "record").lower().replace(" ", "_")
+                )
+                expression_context = {record_variable: record, **field_values}
+
             # Evaluate the expression and set the attribute specified by
             # `self.attribute` to the value it returns.
             try:
-                expression_value = evaluate_expression(expression, names=field_values)
+                expression_value = evaluate_expression(
+                    expression, names=expression_context
+                )
             except simpleeval.NameNotDefined:
                 continue
 
@@ -321,11 +351,11 @@ class FieldType(metaclass=FieldTypeMetaclass):
             custom_applicator = getattr(cls, f"apply_{attribute}", None)
             if custom_applicator:
                 form_field = custom_applicator(
-                    **{
-                        "form_field": form_field,
-                        attribute: expression_value,
-                        "field_values": field_values,
-                    }
+                    form_field,
+                    field=field,
+                    record=record,
+                    field_values=field_values,
+                    **{attribute: expression_value},
                 )
 
             # If no custom applicator method is implemented, but the form field
@@ -349,7 +379,11 @@ class FieldType(metaclass=FieldTypeMetaclass):
 
     @classmethod
     def apply_hidden(
-        cls, form_field: form_fields.Field, hidden: bool = False, **kwargs: Any
+        cls,
+        form_field: form_fields.Field,
+        *,
+        hidden: bool = False,
+        **kwargs: Any,
     ) -> form_fields.Field:
         """Apply the "hidden" attribute.
 
@@ -636,6 +670,13 @@ class ImageUploadField(FieldType):
     model_field_class = ImageField
 
 
+AutocompleteResultMapping = TypedDict(
+    "AutocompleteResultMapping",
+    {"root": str, "value": str, "text": str, "extra": Dict[str, str]},
+    total=False,
+)
+
+
 class AutocompleteSelectField(FieldType):
     """A field for autocompleting selections from an HTTP endpoint."""
 
@@ -645,6 +686,13 @@ class AutocompleteSelectField(FieldType):
     form_widget_class = AutocompleteSelect
     model_field_class = JSONField
 
+    DEFAULT_RESULT_MAPPING: AutocompleteResultMapping = {
+        "root": "@",
+        "value": "id",
+        "text": "text",
+        "extra": {},
+    }
+
     @classmethod
     def as_form_widget(
         cls,
@@ -692,46 +740,28 @@ class AutocompleteSelectField(FieldType):
         )
 
     @classmethod
-    def autocomplete(
+    def render_url(
         cls,
         request: HttpRequest,
         field: "BaseField",
+        url: str,
         **form_widget_options: Any,
-    ) -> Tuple[List[AutocompleteResult], bool]:
-        """Perform a search against the configured URL.
+    ) -> Tuple[str, bool]:
+        """Render the URL using DTL.
 
-        Returns a sequence of dicts, each with "id" and "text" values to use
-        within a select form element.
+        Uses the record (if available) and the request's GET parameters as
+        context when rendering.
 
         Args:
-            request: The HTTP request.
-            field: The Field record instance that uses this autocomplete.
-            form_widget_options: A dict of form widget options that can be
-                used to customize autocomplete behavior.
+            request: The current HttpRequest.
+            field: The BaseField instance.
+            url: The url configured in the form_widget_options.
+            form_widget_options: Options passed to the form widget.
 
         Returns:
-            Tuple[Sequence[AutocompleteResult], bool]: A two-tuple containing
-                a sequence of zero or more select2-compatible search results, and
-                a boolean indicating whether or not there are more results to
-                fetch.
+            Tuple[str, bool]: The rendered URL and a flag indicating whether
+                pagination was handled by the remote endpoint or not.
         """
-        # Extract pagination info.
-        page = int(request.GET.get("page", "1"))
-        per_page = int(request.GET.get("per_page", "100"))
-
-        url = form_widget_options.get("url") or ""
-        results_path = form_widget_options.get("results_path") or "@"
-        text_path = form_widget_options.get("text_path") or "(text || label || name)"
-        id_path = form_widget_options.get("id_path") or "id"
-
-        # Initialize results with an empty list.
-        results: List[AutocompleteResult] = []
-        has_more = False
-
-        # If no URL has been configured, return results as-is.
-        if not url:
-            return results, has_more
-
         # Prevent a circular import.
         from flexible_forms.models import BaseRecord
 
@@ -748,54 +778,195 @@ class AutocompleteSelectField(FieldType):
         # Render the URL as a Django Template, feeding it the current field
         # values, the record instance ("meta"), and the GET parameters.
         url_template = Template(url)
-        rendered_url = url_template.render(
-            Context({"record": record, **request.GET.dict()})
+        record_variable = (
+            (record._meta.verbose_name or "record").lower().replace(" ", "_")
         )
+        url_template_context = Context({record_variable: record, **request.GET.dict()})
 
-        # Get a set of all variable names used in the template so that we can
-        # use them for optimizations later.
-        url_vars = frozenset(
+        # Extract a list of variables referenced in the template.
+        url_template_vars = frozenset(
             v.filter_expression.var.lookups[0]
             for v in url_template.nodelist
             if isinstance(v, VariableNode)
         )
 
-        # Search the endpoint and raise an exception if we get any non-success
-        # status in the response.
-        response = requests.get(rendered_url)
-        response.raise_for_status()
+        # If the "page" variable was used when rendering the URL, assume that
+        # the remote endpoint has handled pagination already.
+        paginated = "page" in url_template_vars
+        filtered = "term" in url_template_vars
 
-        # Decode the JSON response.
-        response_json = response.json()
+        return url_template.render(url_template_context), paginated, filtered
+
+    @classmethod
+    def get_response_json(
+        cls,
+        request: HttpRequest,
+        field: "BaseField",
+        url: str,
+        **form_widget_options: Any,
+    ) -> Any:
+        """Make a request to the configured URL and return the JSON response.
+
+        Args:
+            request: The current HttRequest.
+            field: The BaseField instance.
+            url: The URL from which to fetch results.
+            form_widget_options: Options passed to the form widget.
+
+        Returns:
+            The response from the endpoint, as JSON.
+        """
+        parsed_url = urlparse.urlparse(url)
+        response_json = None
+
+        # If the configured URL is just a path, we might be able to simply
+        # proxy the request to an internal view function instead of making an
+        # entirely new HTTP request. This is less expensive and side-steps
+        # issues with auth when calling local views.
+        if not parsed_url.netloc:
+            try:
+                resolver_match = resolve(parsed_url.path)
+            except Resolver404:  # pragma: no cover
+                pass
+            else:
+                view_func, args, kwargs = resolver_match
+
+                # Invalidate cached properties affected by the request update.
+                for attr, value in type(request).__dict__.items():
+                    if not isinstance(value, cached_property):
+                        continue
+                    try:
+                        delattr(request, attr)
+                    except AttributeError:
+                        pass
+
+                # Update the request to point to the configured URL.
+                request.META["QUERY_STRING"] = parsed_url.query
+                request.GET = QueryDict(parsed_url.query)
+                request.path = parsed_url.path
+                request.path_info = parsed_url.path
+                request.resolver_match = resolver_match
+
+                # Call the view function with the updated request object.
+                response = view_func(request, *args, **kwargs)
+                response_json = json.loads(
+                    getattr(response, "rendered_content", response.content)
+                )
+
+        # If we weren't able to call a local view tog et our results, we'll
+        # make a regular GET request to the URL instead.
+        if response_json is None:
+            response = requests.get(request.build_absolute_uri(url))
+            response.raise_for_status()
+            response_json = response.json()
+
+        return response_json
+
+    @classmethod
+    def extract_results(
+        cls,
+        request: HttpRequest,
+        field: "BaseField",
+        raw_results: Any,
+        mapping: Optional[AutocompleteResultMapping] = None,
+        search_field: Optional[str] = None,
+        search_manually: bool = True,
+        **form_widget_options: Any,
+    ) -> List[AutocompleteResult]:
+        """Extract a list of autocomplete results from the given response JSON.
+
+        Args:
+            request: The current HttpRequest.
+            field: The BaseField instance.
+            raw_results: The raw JSON received from requesting the configured URL.
+            mapping: A mapping of key -> jmespath expression used to
+                translate each result from the URL response into an
+                autocomplete result.
+            form_widget_options: The form_widget_options configured for the field.
+
+        Returns:
+            List[AutocompleteResults]: A list of Select2-compatible autocomplete results.
+        """
+        results = []
+
+        # Define a default result mapping and merge the configured mapping into
+        # it, if given.
+        mapping = cast(
+            AutocompleteResultMapping,
+            {
+                **cls.DEFAULT_RESULT_MAPPING,
+                **(mapping or {}),
+            },
+        )
+
+        search_term = request.GET.get("query", "")
 
         # Parse the search response and map each result to a Select2-compatible
         # dict with an "id" and a "text" property.
-        results_expr = jmespath.compile(results_path)
-        text_expr = jmespath.compile(text_path)
-        id_expr = jmespath.compile(id_path)
-
-        raw_results = results_expr.search(response_json)
+        raw_results = jp(mapping["root"], raw_results) or []
         for result in raw_results:
-            result_text = text_expr.search(result)
-            result_value = id_expr.search(result)
+            result_text = jp(mapping["text"], result)
 
-            if result_value is None:
-                result_value = result_text
+            # Exclude any results that don't contain the search term.
+            if search_manually:
+                searchable_text = str(
+                    result[search_field] if search_field else result_text
+                )
+                if search_term.casefold() not in searchable_text.casefold():
+                    continue
+
+            result_value = jp(mapping["value"], result, default=result_text)
+            result_extra = {k: jp(v, result) for k, v in mapping["extra"].items()}
+
+            result = {
+                "value": result_value,
+                "text": result_text,
+                "extra": result_extra,
+            }
 
             # The "id" is the value eventually stored in the database. In this
             # case, it is a stringified version of the result JSON so that it can
             # be deserialized when rendering the initial value for the widget.
-            result_id = stable_json(
-                {"id": str(result_value), "text": str(result_text), "extra": {}}
-            )
-
+            #
             # The use of "id" as opposed to "value" or something that makes
             # more semantic sense is to support the use of the select2
             # implementation that ships with the Django admin.
-            results.append({"id": result_id, "text": result_text})
+            result["id"] = stable_json(result)
+
+            results.append(result)
+
+        return results
+
+    @classmethod
+    def paginate_results(
+        cls,
+        request: HttpRequest,
+        field: "BaseField",
+        results: List[AutocompleteResult],
+        paginate_manually: bool = True,
+        **form_widget_options: Any,
+    ) -> Tuple[List[AutocompleteResult], bool]:
+        """Paginate the given list of autocomplete results.
+
+        If paginate is True, uses Django's pagination tools to slice the result.
+
+        Args:
+            request: The current HttpRequest.
+            field: The BaseField instance.
+            results: The autocomplete results extracted from the call to the configured URL.
+            paginate: When True, paginates the result set. Otherwise assumes
+                that pagination has been handled already.
+            form_widget_options: The form_widget_options configured for the field.
+
+        Returns:
+            List[AutocompleteResults]: A list of Select2-compatible autocomplete results.
+        """
+        # Extract pagination info from the request.
+        page = int(request.GET.get("page", "1"))
+        per_page = int(request.GET.get("per_page", "100"))
 
         # If the remote endpoint isn't handling pagination, do it manually.
-        if "page" not in url_vars:
+        if paginate_manually:
             paginated_page = Paginator(results, per_page).page(page)
             results, has_more = (
                 cast(List[AutocompleteResult], paginated_page.object_list),
@@ -811,9 +982,156 @@ class AutocompleteSelectField(FieldType):
 
         return results, has_more
 
+    @classmethod
+    def autocomplete(
+        cls,
+        request: HttpRequest,
+        field: "BaseField",
+        url: Optional[str] = "",
+        **form_widget_options: Any,
+    ) -> Tuple[List[AutocompleteResult], bool]:
+        """Perform a search against the configured URL.
+
+        Returns a sequence of dicts, each with "id" and "text" values to use
+        within a select form element.
+
+        Args:
+            request: The HTTP request.
+            field: The Field record instance that uses this autocomplete.
+            url: The URL from which to retrieve autocomplete results.
+            form_widget_options: A dict of form widget options that can be
+                used to customize autocomplete behavior.
+
+        Returns:
+            Tuple[Sequence[AutocompleteResult], bool]: A two-tuple containing
+                a sequence of zero or more select2-compatible search results, and
+                a boolean indicating whether or not there are more results to
+                fetch.
+        """
+        # Initialize results with an empty list.
+        results: List[AutocompleteResult] = []
+        has_more = False
+
+        # If no URL has been configured, return results as-is.
+        if not url:
+            return results, has_more
+
+        rendered_url, paginated, filtered = cls.render_url(
+            request, field, url, **form_widget_options
+        )
+        response_json = cls.get_response_json(
+            request, field, rendered_url, **form_widget_options
+        )
+        results = cls.extract_results(
+            request,
+            field,
+            response_json,
+            search_manually=not filtered,
+            **form_widget_options,
+        )
+
+        return cls.paginate_results(
+            request,
+            field,
+            results,
+            paginate_manually=not paginated,
+            **form_widget_options,
+        )
+
 
 class AutocompleteSelectMultipleField(AutocompleteSelectField):
     """An autocomplete field that supports multiple selections."""
 
     form_field_class = JSONFormField
     form_widget_class = AutocompleteSelectMultiple
+
+
+class QuerysetAutocompleteSelectField(AutocompleteSelectField):
+    """An autocomplete field that sources options from a configured queryset."""
+
+    @classmethod
+    def extract_results(
+        cls,
+        request: HttpRequest,
+        field: "BaseField",
+        raw_results: "QuerySet[DjangoModel]",
+        mapping: Optional[AutocompleteResultMapping],
+        search_field: Optional[str] = None,
+        search_manually: bool = False,
+        **form_widget_options: Any,
+    ) -> List[AutocompleteResult]:
+        mapping = {**cls.DEFAULT_RESULT_MAPPING, **(mapping or {}), "root": "@"}
+
+        # Build a list of model fields referenced in the mapping expressions.
+        expression_fields = set()
+        for expr in (*mapping.values(), *mapping["extra"].values()):
+            if not isinstance(expr, str):
+                continue
+            expression_fields.update(get_expression_fields(expr))
+
+        # Map the model instances to dicts so they can be used with JMESPath.
+        json_results = [
+            {f: getattr(instance, f, None) for f in expression_fields}
+            for instance in raw_results.all()
+        ]
+
+        return super().extract_results(
+            request,
+            field,
+            raw_results=json_results,
+            mapping=mapping,
+            search_field=search_field,
+            search_manually=search_manually,
+            **form_widget_options,
+        )
+
+    @classmethod
+    def autocomplete(
+        cls,
+        request: HttpRequest,
+        field: "BaseField",
+        model: str,
+        filter: Optional[Dict[str, Any]] = None,
+        exclude: Optional[Dict[str, Any]] = None,
+        mapping: Optional[AutocompleteResultMapping] = None,
+        search_field: Optional[str] = None,
+        **form_widget_options: Any,
+    ) -> Tuple[List[AutocompleteResult], bool]:
+        # Build the queryset from the configured parameters.
+        model_cls = cast(DjangoModel, apps.get_model(model))
+        qs = (
+            model_cls._default_manager.filter(**(filter or {}))
+            .exclude(**(exclude or {}))
+            .order_by("pk")
+        )
+
+        # If the field is configured with a search_field, use it to filter results.
+        search_term = request.GET.get("term")
+        if search_term and search_field:
+            qs = qs.filter(**{f"{search_field}__icontains": search_term})
+
+        # Paginate the results first (to allow Django's paginator to work with the QuerySet).
+        results, has_more = cls.paginate_results(
+            request, field, results=qs, paginate_manually=True, **form_widget_options
+        )
+
+        # Map the results to options using the given mapping.
+        results = cls.extract_results(
+            request,
+            field,
+            raw_results=results,
+            mapping={
+                **(mapping or {}),
+                "root": "@",
+            },
+            search=False,
+        )
+
+        return results, has_more
+
+
+class QuerysetAutocompleteSelectMultipleField(QuerysetAutocompleteSelectField):
+    """A queryset autocomplete field that supports multiple selections."""
+
+    form_field_class = JSONFormField
+    form_widget_class = AutocompleteSelectMultiple
